/*  Pcsx2 - Pc Ps2 Emulator
*  Copyright (C) 2009  Pcsx2-Playground Team
*
*  This program is free software; you can redistribute it and/or modify
*  it under the terms of the GNU General Public License as published by
*  the Free Software Foundation; either version 2 of the License, or
*  (at your option) any later version.
*  
*  This program is distributed in the hope that it will be useful,
*  but WITHOUT ANY WARRANTY; without even the implied warranty of
*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*  GNU General Public License for more details.
*  
*  You should have received a copy of the GNU General Public License
*  along with this program; if not, write to the Free Software
*  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
*/

#pragma once
#ifdef PCSX2_MICROVU

//------------------------------------------------------------------
// Micro VU - Clamp Functions
//------------------------------------------------------------------

// Used for Result Clamping
microVUx(void) mVUclamp1(int reg, int regTemp, int xyzw) {
}

// Used for Operand Clamping
microVUx(void) mVUclamp2(int reg, int regTemp, int xyzw) {
}

//------------------------------------------------------------------
// Micro VU - Misc Functions
//------------------------------------------------------------------

microVUx(void) mVUunpack_xyzw(int dstreg, int srcreg, int xyzw) {
	switch ( xyzw ) {
		case 0: SSE2_PSHUFD_XMM_to_XMM(dstreg, srcreg, 0x00); break;
		case 1: SSE2_PSHUFD_XMM_to_XMM(dstreg, srcreg, 0x55); break;
		case 2: SSE2_PSHUFD_XMM_to_XMM(dstreg, srcreg, 0xaa); break;
		case 3: SSE2_PSHUFD_XMM_to_XMM(dstreg, srcreg, 0xff); break;
	}
}

microVUx(void) mVUloadReg(int reg, u32 offset, int xyzw) {
	switch( xyzw ) {
		case 8:		SSE_MOVSS_M32_to_XMM(reg, offset);		break; // X
		case 4:		SSE_MOVSS_M32_to_XMM(reg, offset+4);	break; // Y
		case 2:		SSE_MOVSS_M32_to_XMM(reg, offset+8);	break; // Z
		case 1:		SSE_MOVSS_M32_to_XMM(reg, offset+12);	break; // W
		case 3:		SSE_MOVHPS_M64_to_XMM(reg, offset+8);	break; // ZW (not sure if this is faster than default)
		case 12:	SSE_MOVLPS_M64_to_XMM(reg, offset);		break; // XY (not sure if this is faster than default)
		default:	SSE_MOVAPS_M128_to_XMM(reg, offset);	break;
	}
}

microVUx(void) mVUsaveReg(int reg, u32 offset, int xyzw) {
	switch ( xyzw ) {
		case 1: // W
			SSE_MOVSS_XMM_to_M32(offset+12, reg);
			break;
		case 2: // Z
			SSE_MOVSS_XMM_to_M32(offset+8, reg);
			break;
		case 3: // ZW
			SSE_MOVHPS_XMM_to_M64(offset+8, reg);
			break;
		case 4: // Y
			SSE_MOVSS_XMM_to_M32(offset+4, reg);
			break;
		case 5: // YW
			SSE_SHUFPS_XMM_to_XMM(reg, reg, 0xB1);
			SSE_MOVHLPS_XMM_to_XMM(xmmT1, reg);
			SSE_MOVSS_XMM_to_M32(offset+4, reg);
			SSE_MOVSS_XMM_to_M32(offset+12, xmmT1);
			break;
		case 6: // YZ
			SSE2_PSHUFD_XMM_to_XMM(xmmT1, reg, 0xc9);
			SSE_MOVLPS_XMM_to_M64(offset+4, xmmT1);
			break;
		case 7: // YZW
			SSE2_PSHUFD_XMM_to_XMM(xmmT1, reg, 0x93); //ZYXW
			SSE_MOVHPS_XMM_to_M64(offset+4, xmmT1);
			SSE_MOVSS_XMM_to_M32(offset+12, xmmT1);
			break;
		case 8: // X
			SSE_MOVSS_XMM_to_M32(offset, reg);
			break;
		case 9: // XW
			SSE_MOVHLPS_XMM_to_XMM(xmmT1, reg);
			SSE_MOVSS_XMM_to_M32(offset, reg);
			if ( cpucaps.hasStreamingSIMD3Extensions ) SSE3_MOVSLDUP_XMM_to_XMM(xmmT1, xmmT1);
			else SSE_SHUFPS_XMM_to_XMM(xmmT1, xmmT1, 0x55);
			SSE_MOVSS_XMM_to_M32(offset+12, xmmT1);
			break;
		case 10: //XZ
			SSE_MOVHLPS_XMM_to_XMM(xmmT1, reg);
			SSE_MOVSS_XMM_to_M32(offset, reg);
			SSE_MOVSS_XMM_to_M32(offset+8, xmmT1);
			break;
		case 11: //XZW
			SSE_MOVSS_XMM_to_M32(offset, reg);
			SSE_MOVHPS_XMM_to_M64(offset+8, reg);
			break;
		case 12: // XY
			SSE_MOVLPS_XMM_to_M64(offset, reg);
			break;
		case 13: // XYW
			SSE2_PSHUFD_XMM_to_XMM(xmmT1, reg, 0x4b); //YXZW				
			SSE_MOVHPS_XMM_to_M64(offset, xmmT1);
			SSE_MOVSS_XMM_to_M32(offset+12, xmmT1);
			break;
		case 14: // XYZ
			SSE_MOVHLPS_XMM_to_XMM(xmmT1, reg);
			SSE_MOVLPS_XMM_to_M64(offset, reg);
			SSE_MOVSS_XMM_to_M32(offset+8, xmmT1);
			break;
		case 15: // XYZW				
			SSE_MOVAPS_XMM_to_M128(offset, reg);
			break;
	}
}

#endif //PCSX2_MICROVU